#!/usr/bin/python
import os, os.path, sys, re, glob, itertools

########################################################
################# READING
def readParameters(fp=None):
    tf_dict = {"True":True, "False":False, "true":True, "false": False}
    noparse = ["externals", "ks", "kins", "lmbds", "mrlmbds", "csts"]

    ### pickle: -1 disabled, 0 store, 1 load
    ### log: - stdout, + auto filename, * both, other filename
    ### normalize comp: flip comparison edges A>B to A<B
    ### try simply: try finding simpler clauses when remapping 
    ### try V: which shapes to try 1 /\, 2 \/, 3 -- positive indices: even if simple edge has enough support
    org_parameters = {'verbosity': 3, 'logfile': "+", "ego": -1, "center":"", "min_supp": 3,
                      'results_rep': "../results/", "data_rep":"../data/", 'graph_ext': "graph",
                      "min_count": 1.0, "action": "dos", "lmbd": 1.0,
                      "trans": 0, "ks":"0", "output": "+", "min_weight": None, "type_support":"int",
                      "adjust":1}

    # 'limit_paths_itemsets': 1000, 'limit_graphs_itemsets': 1000,
    # 'limit_unit_instances': 0, 'limit_abs_instances': 10**5,

    if fp is not None:
        read_parameters = {}
        for line in fp:
            if re.match("^ *#", line) is not None:
                continue
            parts = line.split()
            if len(parts) > 1:
                if parts[0] in noparse:
                    tmp = " ".join(parts[1:])
                    read_parameters[parts[0]] = tmp
                elif parts[1] in tf_dict:
                    read_parameters[parts[0]] = tf_dict[parts[1]]
                else:
                    try:
                        read_parameters[parts[0]] = int(parts[1])
                    except ValueError:
                        try:
                            read_parameters[parts[0]] = float(parts[1])
                        except ValueError:
                            if "~" in parts[1]:
                                tmp = re.sub("~", os.path.expanduser("~"), parts[1])
                            else:
                                tmp = " ".join(parts[1:])
                            read_parameters[parts[0]] = tmp
        org_parameters.update(read_parameters)

    org_parameters["ks"] = [int(s) for s in org_parameters["ks"].strip().split(",")]
    if "kins" in org_parameters:
        org_parameters["kins"] = [int(s) for s in org_parameters["kins"].strip().split(",")]
    if "lmbds" in org_parameters:
        org_parameters["lmbds"] = [float(s) for s in org_parameters["lmbds"].strip().split(",")]
    if "mrlmbds" in org_parameters:
        org_parameters["mrlmbds"] = [float(s) for s in org_parameters["mrlmbds"].strip().split(",")]

    return org_parameters

def readGraph(filename, min_weight=None, center=None):
    graph = {}
    # with codecs.open(filename, encoding='utf-8', mode='r') as f:
    sep = None
    counts = 0
    with open(filename, mode='r') as f:
        for line in f:
            if re.match("\s*#", line):
                continue
            # #### WARNING ONLY DEBUG!
            # counts += 1
            # if counts > 1000:
            #     break

            if sep is None:
                tmp = re.match("^(?P<from>[^\t]*)\t(?P<to>[^\t]*)(\t(?P<count>[0-9.]*) *)?$", line.strip())
                if tmp is None:
                    sep = " "                    
                else:
                    sep = "\t"
            tmp = re.match("^(?P<from>[^"+sep+"]*)"+sep+"(?P<to>[^"+sep+"]*)("+sep+"(?P<count>[0-9.]*) *)?$", line.strip())
            if tmp is not None and tmp.group("from") != tmp.group("to"):
                nfrom, nto, ncount = tmp.group("from"), tmp.group("to"), None
                if tmp.group("count") is not None:
                    ncount = float(tmp.group("count"))
                if ncount >= min_weight and center not in [nfrom, nto]:
                    for (f,t) in [(nfrom, nto), (nto, nfrom)]:
                        if not graph.has_key(f):
                            graph[f] = {t:ncount}
                        else:
                            graph[f][t] = ncount
    return graph

def cleanCache(data_parameters):
    for v in ["store_utils"]:
        if data_parameters.has_key(v):
            del data_parameters[v]

########################################################
################# PRINTING
def openMakeP(filename, mode="w"):
    d = os.path.dirname(filename)
    if not os.path.isdir(d):
        os.makedirs(d)
    return open(filename, mode)

def initResOut(data_parameters, external, k):
    if data_parameters["output"] != "-":
        if data_parameters["output"] == "+":
            fbasis = "%s%s.%s_%d" % (data_parameters['results_rep'], data_parameters['basis'], external, k)
            fo = openMakeP(fbasis+".communities", "w")
            # fe =None
            fe = openMakeP(fbasis+".cclust", "w")
        else:
            fbasis = data_parameters["output"]
            fo = openMakeP(fbasis+".communities", "w")
            fe = fo
    else:
        fo = sys.stdout
        fe = None
    fo.write("------------METHOD=%s K=%d\n" % (external, k))
    return fo, fe

class Dummyfp:
    def write(self, txt):
        pass
    def close(self):
        pass
    
def initResOutSimple(data_parameters, external, k):
    if data_parameters.get("output", "-") == "-":
        return Dummyfp(), Dummyfp()
    if data_parameters.get("output", "-") != "-":
        if data_parameters["output"] == "+":
            fbasis = "%s%s.%s_%d" % (data_parameters['results_rep'], data_parameters['basis'], external, k)
            fo = openMakeP(fbasis+".communities", "w")
            fe = openMakeP(fbasis+".cclust", "w")
        else:
            fbasis = data_parameters["output"]
            fo = openMakeP(fbasis+".communities", "w")
            fe = fo

    else:
        fo = sys.stdout
        fe = None
    fo.write("------------METHOD=%s K=%d lmbd=%f\n" % (external, k, data_parameters["lmbd"]))
    return fo, fe

def closeOut(fo):
    if fo is not None and fo != sys.stdout:
        fo.close()


########################################################
################# MISC
def filterGBBox(graph, ndloc_dict, data_parameters):
    bboxes = {"USA" : ((48.504703, -128.360312),
                       (23.628067, -61.915000)),
              "EU": ((66.365554, -27.989219),
                     (36.300649, 35.116250)),
              "East-Asia" : ((-8.772548, 91.542031),
                             (50.725084, 154.823281)),
              "Au-NZ": ((-50.412964, 180),
                        (-9.466782, 91.542031)),
              "Middle-East": ((42.523863, 24.433596),
                              (10.102579, 68.466398)),
              "Latin-America":
              ((27.180287, -119.003904),
               (-55.502634, -34.632139)),
              "Sub-Sahara":
              ((19.107400, -21.709384),
               (-36.770911, 57.743742)),
              "India":
              ((3.297826, 93.545822),
               (32.734995, 60.411056))
              }

    #### COORDS GIVEN arbitrary (LAT, LONG) corners turn to (min, max) in  (LAT, LONG)
    for b in bboxes.keys():
        bboxes[b] = ((min(bboxes[b][0][0], bboxes[b][1][0]),
                      min(bboxes[b][0][1], bboxes[b][1][1])),
                     (max(bboxes[b][0][0], bboxes[b][1][0]),
                      max(bboxes[b][0][1], bboxes[b][1][1])))

    nodesets = dict([(bb, set()) for bb in bboxes.keys()])
    for b, loc in ndloc_dict.items():
        if loc != (0,0):
            for bbox, corners in bboxes.items():
                if loc[0] > corners[0][0] and loc[0] < corners[1][0] and \
                       loc[1] > corners[0][1] and loc[1] < corners[1][1]:
                    nodesets[bbox].add(b)

    for bbox, nds in nodesets.items():
        fo = openMakeP(data_parameters['data_rep'] + ("dbbox_%s.edges" % bbox), "w")
        fol = openMakeP(data_parameters['data_rep'] + ("dbbox_%s.locations" % bbox), "w")
        for node in nds:
            fol.write("%s\t%f\t%f\n" % (node, ndloc_dict[node][0], ndloc_dict[node][1]))
            for nt in nds.intersection(graph.get(node, {}).keys()):
                fo.write("%s\t%s\n" % (node, nt))
        fo.close()
        fol.close()


