import copy, datetime, itertools, numpy, subprocess, glob
from sortedcontainers import SortedDict
import scipy.sparse as spm

from experimental import *

EXP=-1
THRESD=4*10**4

def comms_to_mat(comms, map_nd, Cm=None):
    if Cm is not None and len(comms) == Cm.shape[1]:
        return Cm
    off = 0
    if Cm is not None:
        off = Cm.shape[1]
    N = len(map_nd)
    M = len(comms)-off
    if N > 10**EXP:
        adj = spm.lil_matrix((N, M))
    else:
        adj = numpy.zeros((N, M))
    for ci, cc in enumerate(comms[off:]):
        idx = [map_nd[n] for n in cc]
        adj[idx,ci] = 1.
    if Cm is not None:
        if N > 10**EXP:
            adj = spm.hstack([Cm, adj])
        else:
            adj = numpy.hstack([Cm, adj])
    if N > 10**EXP:
        adj = spm.csc_matrix(adj)

    return adj

def graph_to_map(graph):
    comm =  sorted(graph.keys())
    return dict([(v,k) for (k,v) in enumerate(comm)])

def graph_to_mat(graph, map_nd):
    N = len(map_nd)
    if N > 10**EXP:
        adj = spm.lil_matrix((N, N))
    else:
        adj = numpy.zeros((N, N))

    for node, neighs in graph.items():
        adj[map_nd[node], [map_nd[neigh] for neigh in neighs.keys()]] = 1.
    if N > 10**EXP:
        adj = spm.csc_matrix(adj)

    return adj

def dist_sets(X,Y):
    lX, lY = len(X), len(Y)
    I = len(X.intersection(Y))
    return dist_l(lX,lY,lI)

def dist_l(lX,lY,lI):
    if lX == lI and lY==lI:
        return 0.
    else:
        return 2 - lI**2./(lX*lY)

def find_top_n_bycJ(comm_groups, n_to_sqrtci, lmbd, ci_counts, ti):
    cks = comm_groups.keys()
    top = (float('Inf'), None, None)
    for ck in cks:
        if len(comm_groups[ck]) > 0:
            cnd, cv = comm_groups[ck].popitem()
            comm_groups[ck][cnd] = cv ### PUTTING IT RIGHT BACK, that was just for comparison
            # cnd = comm_groups[ck].keys()[0]
            sc = -cnd[0]-4*lmbd*sum([ci_counts[ci][0]/ci_counts[ci][1] for ci in ck])
            #### if the score is lower or it's equal and the node id is lower
            if sc < top[0] or sc == top[0] and cnd[1] > top[1]:
                top = (sc, cnd[1], -cnd[0])
    return top

def update_cidx(n_to_cidx, dense, ti):
    for n in dense:
        n_to_cidx[n] = n_to_cidx.get(n, ()) + (ti,) 
    return ti

def update_cgJ(cg, n, graph, current_degrees, n_to_cidx):
    #### for each neighbor of the removed node,
    #### degree has decreased so need to update the position accordingly
    ncount = 0
    for ne in graph[n].keys():
        if current_degrees.get(ne,0) > 0:
            ncount +=1
            v = n_to_cidx.get(ne,())
            nv = cg[v].pop((-current_degrees[ne], ne))
            current_degrees[ne] -= 1
            cg[v][(-current_degrees[ne], ne)] = nv
    return ncount

def update_scores_c(neighs, scores):
    ### going over the nodes in the community group, popping the neighbors,
    ### decreasing their score and reinserting them in their new position
    ### the order within the neighbors is maintained since they all decrease degree by one
    ordered_neighs = []
    i = 0
    while len(neighs)+len(ordered_neighs) > 0 and i < len(scores):
        if scores[i][1] in neighs:
            neighs.remove(scores[i][1])
            tmp = scores.pop(i)
            ordered_neighs.append((tmp[0]-1, tmp[1]))
        #### :SORT: elif len(ordered_neighs) > 0 and scores[i] < ordered_neighs[0]:
        elif len(ordered_neighs) > 0 and (-scores[i][0],scores[i][1]) > (-ordered_neighs[0][0], ordered_neighs[0][1]):
            ### sorted by degrees first, then by labels
            scores.insert(i, ordered_neighs.pop(0))
            i += 1
        else:
            i += 1
    ### in case there remain some nodes
    scores.extend(ordered_neighs)

def prepareSeriesLmbd(action, data_parameters):
    if "lmbds" in data_parameters:
        if action in ["overn"]:
            return [{"lmbd":0.0, "elmbd": "%s" % 0}], None
        if action in ["dos"]:
            return [{"lmbd":v, "elmbd": "%s" % v} for v in data_parameters["lmbds"]], None
    return [{}], None

def getFrontDOS(graph, data_parameters):
    Nl = -8
    data_parameters['obasis'] = data_parameters['basis']
    data_parameters['basis'] = data_parameters['obasis'] + ("_l%s-%s" % (Nl,  data_parameters["serie_suff"]))
    details, comms, fingpN = findDensestOverN(graph, data_parameters)
    dtp = dict(data_parameters)
    dtp["ks"] = [1]
    dtp["lmbd"] = 0.
    data_parameters['basis'] = data_parameters['obasis'] + ("_l%s-%s" % (0,  data_parameters["serie_suff"]))
    details, comms, fingp = findDensestOverN(graph, dtp)
    outs = {Nl: fingpN, 0: fingp}
    getFrontBnds(graph, data_parameters, [Nl, 0], outs, 0, 15)

def getFrontBnds(graph, data_parameters, lmbds, outs, rec=0, maxrec=None):
    if maxrec is None or rec < maxrec:
        clmbd = lmbds[0]/2.0
        data_parameters["lmbd"] = clmbd
        data_parameters['basis'] = data_parameters['obasis'] + ("_l%s-%s" % (data_parameters["lmbd"],  data_parameters["serie_suff"]))
        details, comms, fingp = findDensestDOS(graph, data_parameters)
        print ">>>", clmbd, fingp
        if fingp != outs[lmbds[0]] and fingp != outs[lmbds[1]]:
            outs[clmbd] = fingp
            getFrontBnds(graph, data_parameters, [clmbd, lmbds[1]], outs, rec+1, maxrec)


### Denset subgraph with overlap, main algorithm
def findDensestDOS(graph, data_parameters):

    min_supp = 0 #data_parameters["min_supp"] #*(data_parameters["min_supp"]-1)
    lmbd = data_parameters["lmbd"] #*(data_parameters["min_supp"]-1)
    ks = data_parameters["ks"]
    external = "DOS"
    kmax = 0

    if data_parameters.get("adjust", 0) == 0:
         external += "-a0"

    if 0 not in ks:
        kmax = max(ks)
    print "------------METHOD=%s K=%d lmbd=%f\n" % (external, kmax, lmbd)

    fbasis = "%s%s.%s_%d" % (data_parameters['results_rep'], data_parameters['basis'], external, kmax)
    # if os.path.isfile(fbasis + ".cclust"):
    #    print "I DID THAT ALREADY, SKIPPING..." 
    #    return

    fo, fe = initResOutSimple(data_parameters, external, kmax)
    ticE = datetime.datetime.now()

    org_graph = graph

    curr_degrees = dict([(n, len(v)) for n,v in org_graph.items()])    
    
    totnb_nodes = float(len(curr_degrees))
    stop = False    
    ti = -1

    n_to_cidx = {}
    comm_groups = {(): SortedDict([((-d,n), n) for n,d in curr_degrees.items()])}


    comms = []
    ticcE = datetime.datetime.now()
    details = []
    adjusted_counts = {-2: 0, -1: 0, 1:0, 0:0, 100:0}
    initdegrees = sum(curr_degrees.values())

    while not stop:
        ### prepare scores
        ###
        ti += 1
        history = []

        ssdegrees = initdegrees
        ci_counts = [[d[0], d[0]] for d in comms] 
        nb_nodes = totnb_nodes
        n = None                
        while nb_nodes > 0:
            history.append((ssdegrees/(2.*nb_nodes)+lmbd*(2*ti - sum([ccc[0]**2/ccc[1] for ccc in ci_counts])/nb_nodes), n))
            sc, n, dd = find_top_n_bycJ(comm_groups, n_to_cidx, lmbd, ci_counts, ti)
            if nb_nodes % 10000 == 0:
                print "---", nb_nodes, dd, datetime.datetime.now()-ticcE
                ticcE = datetime.datetime.now()
            ### it's the greatest -degree of it's group...
            comm_groups[n_to_cidx.get(n,())].popitem()
            ncount = update_cgJ(comm_groups, n, org_graph, curr_degrees, n_to_cidx)
    
            ssdegrees -= 2*ncount
            for c in n_to_cidx.get(n, ()):
                ci_counts[c][0] -= 1 
            del curr_degrees[n]

            nb_nodes -= 1

        history.append((float("-Inf"), n))
        for n in curr_degrees.keys():
            history.append((float("-Inf"), n))
        t = max([(v[0],-i) for i,v in enumerate(history)])
        print t

        if len(history)+t[1]-1 < min_supp or t[0] == float("-Inf"):
            stop = True
            pdb.set_trace()
        else:
            dense = set([h[1] for h in history[-t[1]+1:]])
            tacE = datetime.datetime.now()
            
            if data_parameters.get("adjust", 0) == 1:
                dense, dscore, avg_degree, adjsing = comm_adjust(org_graph, dense, comms, lmbd, n_to_cidx)
                adjusted_counts[adjsing] +=1
            else:
                avg_degree = sum([len(dense.intersection(org_graph[n])) for n in dense])/(2.0*len(dense))

            if ti > 0:
                int_sizes = [len(dense.intersection(c[-1])) for c in comms]
                jaccs = [int_sizes[ccii]/float(len(dense.union(c[-1]))) for ccii, c in enumerate(comms)]
                dists = [dist_l(len(dense), len(c[-1]), int_sizes[ccii]) for ccii, c in enumerate(comms)]
                dist = sum(dists)
            else:
                int_sizes = []
                jaccs = []
                dists = []
                dist = 0.0

            #### keys of n_to_cidx are nodes that already belong to some community
            nb_new = len(dense.difference(n_to_cidx.keys()))
            tot = avg_degree+lmbd*dist

            details.append({"ti": ti, "avg_degree": avg_degree, "dist": dist, "lmbd": lmbd, "nodes": dense, "nb_nodes": len(dense), "int_sizes": int_sizes, "jaccs": jaccs, "dists": dists, "nb_new": nb_new, "tot": tot, "timer": (tacE - ticE)})

            if ti == 0 and lmbd < 0:
                ### SET LAMBDA as a function of the size of first community found
                lmbd *= -details[-1]["tot"]

            ### update with the new found community
            ###
            update_cidx(n_to_cidx, dense, ti)
            curr_degrees = dict([(n, len(v)) for n,v in org_graph.items()])    
            comm_groups = {(): []}
            for n in curr_degrees.keys():
                v = n_to_cidx.get(n, ())
                if v in comm_groups:
                    comm_groups[v].append(((-curr_degrees[n],n), n))
                else:
                    comm_groups[v]= [((-curr_degrees[n],n), n)]
            for v in comm_groups.keys():
                comm_groups[v] = SortedDict(comm_groups[v])
            comms.append([float(len(dense)), dense])

            if fo is not None:
                fo.write("Here in %s\n" % (tacE - ticE))
                fo.write("* (%(ti)d)\ttot=%(tot)f avg_degree=%(avg_degree)f dist=%(dist)f lmbd=%(lmbd)f nb_nodes=%(nb_nodes)d nb_new=%(nb_new)d int_sizes=%(int_sizes)s jaccs=%(jaccs)s dists=%(dists)s\n" % details[-1])
            if fe is not None:
                fe.write("%s\n" % (", ".join(dense)))

        if 0 not in ks and ti+1 >= max(ks):
            stop = True

            jaccs = list(itertools.chain(*[d["jaccs"] for d in details]))

    tot_time = datetime.datetime.now() - ticE
    printOut(fo, tot_time, lmbd, totnb_nodes, details, graph)
    closeOut(fo)
    closeOut(fe)
    if data_parameters.get("adjust", 0) == 1:
        print "ADJUSTED:", adjusted_counts
    fingp = fingerprint(details)
    return details, comms, fingp        

### Adjusting the community if the current has already been found
def comm_adjust(graph, dense, comms, lmbd, n_to_cidx):
    nbedges_org = sum([len(dense.intersection(graph[n])) for n in dense])
    nbnodes_org = 1.0*len(dense)
    avg_degree = nbedges_org/(2*nbnodes_org)

    ti = len(comms)
    ci_counts = [(len(dense.intersection(comm[1])), comm[0]) for comm in comms]
    score_org = avg_degree+lmbd*(2*ti - sum([ccc[0]**2/ccc[1] for ccc in ci_counts])/nbnodes_org)
    
    seen = comm_seen(dense, comms)
    best = [score_org, avg_degree, None, 0]
    if len(seen) > 0:
        best = [0, 0, None, 100]

        non_cands_add = comm_one_more(dense, comms)
        cands_add = set(graph.keys()).difference(dense, non_cands_add)
        if len(cands_add) == 0 and avg_degree <= 7./6:
            ### deduct overlap penalty from avg_degree
            trip = best_unseen_triple(graph, comms, lmbd, n_to_cidx)
            return trip[-1], trip[1], trip[0], -2

        if len(dense) > 1:
            for cand in dense.difference(comm_one_less(dense, comms)):
                avg_degree_new = (nbedges_org-2.*len(dense.intersection(graph[cand])))/(2*(nbnodes_org-1))
                score_new = avg_degree_new+lmbd*(2*ti - sum([(ccc[0]-(ccci in n_to_cidx.get(cand,[])))**2/ccc[1] for ccci, ccc in enumerate(ci_counts)])/(nbnodes_org-1))
                if score_new > best[0] or (score_new == best[0] and cand < best[1]):
                    best = [score_new, avg_degree_new, cand, -1]
            
        for cand in cands_add:
            avg_degree_new = (nbedges_org+2.*len(dense.intersection(graph[cand])))/(2*(nbnodes_org+1))
            score_new = avg_degree_new+lmbd*(2*ti - sum([(ccc[0]+(ccci in n_to_cidx.get(cand,[])))**2/ccc[1] for ccci, ccc in enumerate(ci_counts)])/(nbnodes_org+1))
            if score_new > best[0] or (score_new == best[0] and cand < best[1]):
                best = [score_new, avg_degree_new, cand, 1]

    if best[-1] == 1:
        dense.add(best[-2])
    elif best[-1] == -1:
        dense.remove(best[-2])
    return dense, best[0], best[1], best[-1]



def best_unseen_triple(graph, comms, lmbd, n_to_cidx):
    print "Really trying triples. Really! How did I got to such extremity!!.."
    best = [0.0, 0.0, None]
    seen_trips = [set(comms[1]) for ci, comm in enumerate(comms) if len(comms[1]) == 3]
    for trip in map(set, itertools.combinations(graph.keys(), 3)):
        if trip not in seen_trips:
            deg = sum([len(trip.intersection(graph[n])) for n in trip])/6.
            if deg > best[0]:
                ci_counts = [(len(dense.intersection(comm[1])), comm[0]) for comm in comms]
                score = deg+lmbd*(2*len(comms) - sum([len(trip.intersection(comm[1]))**2/comm[0] for comm in comms])/3.)
                if score > best[1]:
                    best = [deg, score, trip]
    return best

### find already seen subgraph that match dense
def comm_seen(dense, comms):
    return [ci for ci, comm in enumerate(comms) if len(dense) == comm[0] and dense == comm[1]]

### find vertices not in dense such that when added to dense the result is an already seen subgraph
def comm_one_more(dense, comms):
    nds = set()
    for ci, comm in enumerate(comms):
        if len(dense) == comm[0]-1:
            d = comm[1].difference(dense)
            if len(d) == 1:
                nds.add(d.pop())
    return nds

### find vertices from dense such that when removed from dense the result is an already seen subgraph
def comm_one_less(dense, comms):
    nds = set()
    for ci, comm in enumerate(comms):
        if len(dense) == comm[0]+1:
            d = dense.difference(comm[1])
            if len(d) == 1:
                nds.add(d.pop())
    return nds


### Denset subgraph without any overlap,
###  just remove communities as they are found,
###  ought to be faster than setting lmbd=0 in main algo
def findDensestOverN(graph, data_parameters):
    ##### TEST WITH SORTED LIST....
    min_supp = 0 #data_parameters["min_supp"] #*(data_parameters["min_supp"]-1)
    lmbd = data_parameters["lmbd"] #*(data_parameters["min_supp"]-1)
    ks = data_parameters["ks"]
    external = "overN"
    kmax = 0
    lmbd = 0.0
    
    if data_parameters.get("drop_ones", 0) == 1:
         external += "-dO"

    if 0 not in ks:
        kmax = max(ks)
    print "------------METHOD=%s K=%d lmbd=%f\n" % (external, kmax, lmbd)

    fo, fe = initResOutSimple(data_parameters, external, kmax)
    ticE = datetime.datetime.now()

    if data_parameters.get("drop_ones", 0) == 1:
        ones = []
        org_graph = {}
        for n,v in graph.items():
            if len(v) > 1:
                org_graph[n] = copy.deepcopy(v)
            else:
                ones.append(n)

        for n in ones:
            try:
                del org_graph[graph[n].keys()[0]][n]
            except KeyError:
                pass
    else:
        org_graph = graph

    curr_degrees = dict([(n, len(v)) for n,v in org_graph.items()])    
    
    totnb_nodes = float(len(curr_degrees))
    stop = False    
    ti = -1
    unions = set(org_graph.keys())
    queue = SortedDict([((-d,n), n) for n,d in curr_degrees.items()])

    comms = []
    ticcE = datetime.datetime.now()
    details = []

    while not stop:
        ### prepare scores
        ###
        ti += 1
        history = []

        ssdegrees = sum(curr_degrees.values())
        ci_counts = [[d[0], d[0]] for d in comms] 
        nb_nodes = len(curr_degrees)
        n = None                
        while len(queue) > 0:
            history.append((ssdegrees/(2.*nb_nodes), n))
            cnd, cv = queue.popitem()
            n, dd = cnd[1], -cnd[0]
            if nb_nodes % 10000 == 0:
                print "---", nb_nodes, dd, datetime.datetime.now()-ticcE
                ticcE = datetime.datetime.now()
            ncount = 0
            for ne in graph[n].keys():
                if curr_degrees.get(ne,0) > 0:
                    ncount +=1
                    nv = queue.pop((-curr_degrees[ne], ne))
                    curr_degrees[ne] -= 1
                    queue[(-curr_degrees[ne], ne)] = nv

            ssdegrees -= 2*ncount
            del curr_degrees[n]
            nb_nodes -= 1

        history.append((float("-Inf"), n))
        for n in curr_degrees.keys():
            history.append((float("-Inf"), n))
        t = max([(v[0],-i) for i,v in enumerate(history)])

        print t
        if len(history)+t[1]-1 < min_supp or t[1] == 0 or t[0] == float("-Inf"):
            stop = True
        else:
            dense = set([h[1] for h in history[-t[1]+1:]])
            tacE = datetime.datetime.now()
            
            avg_degree = sum([len(dense.intersection(org_graph[n])) for n in dense])/(2.0*len(dense))
            if ti > 0:
                int_sizes = [len(dense.intersection(c[-1])) for c in comms]
                jaccs = [0.0 for ccii, c in enumerate(comms)]
                dists = [2.0 for ccii, c in enumerate(comms)]
                dist = sum(dists)
            else:
                int_sizes = []
                jaccs = []
                dists = []
                dist = 0.0

            #### keys of n_to_cidx are nodes that already belong to some community
            nb_new = len(dense)
            tot = avg_degree
            details.append({"ti": ti, "avg_degree": avg_degree, "dist": dist, "lmbd": lmbd, "nodes": dense, "nb_nodes": len(dense), "int_sizes": int_sizes, "jaccs": jaccs, "dists": dists, "nb_new": nb_new, "tot": tot, "timer": (tacE - ticE)})

            ### update with the new found community
            ###
            comms.append([float(len(dense)), dense])
            unions.difference_update(dense)
            curr_degrees = dict([(n, len(unions.intersection(graph[n].keys()))) for n in unions])
            queue = SortedDict([((-d,n), n) for n,d in curr_degrees.items()])

            if fo is not None:
                fo.write("Here in %s\n" % (tacE - ticE))
                fo.write("* (%(ti)d)\ttot=%(tot)f avg_degree=%(avg_degree)f dist=%(dist)f lmbd=%(lmbd)f nb_nodes=%(nb_nodes)d nb_new=%(nb_new)d int_sizes=%(int_sizes)s jaccs=%(jaccs)s dists=%(dists)s\n" % details[-1])
            if fe is not None:
                fe.write("%s\n" % (", ".join(dense)))

        if 0 not in ks and ti+1 >= max(ks):
            stop = True
            
    tot_time = datetime.datetime.now() - ticE
    printOut(fo, tot_time, lmbd, totnb_nodes, details, graph)
    closeOut(fo)
    closeOut(fe)
    fingp = fingerprint(details)
    return details, comms, fingp        

def printOut(fo, tot_time, lmbd, totnb_nodes, details, graph=None, data={}):

    if fo is not None:
        jaccs = list(itertools.chain(*[d["jaccs"] for d in details]))
        int_sizes = list(itertools.chain(*[d["int_sizes"] for d in details]))
        if len(details) == 1:
            jaccs = [0]
            int_sizes = [0]
        c_sizes = [d["nb_nodes"] for d in details]

       #### FOR COMPUTING MODULARITY
        Am = None
        map_nd = None
        Cm = None
        mods = {}
        if "Am" in data:
            Am = data["Am"]
        elif graph is not None:
            map_nd = graph_to_map(graph)
            Am = graph_to_mat(graph, map_nd)
        
        if "Cm" in data:
            Cm = data["Cm"]
        elif graph is not None:
            if map_nd is None:
                map_nd = graph_to_map(graph)
            Cm = comms_to_mat([d["nodes"] for d in details], map_nd)

        if Am is not None and Cm is not None:
            mods = modularity(Am, Cm, ["E", "C"])
        if Cm is not None:
            multipln = Cm.sum(1)
            multipln = numpy.array(multipln[multipln>0])
            nbcover = multipln.shape[1]
            if len(mods) == 0 and "map_nd" in data:
                mods = modularityE(graph, data["map_nd"], Cm, [d["nodes"] for d in details])
        else:
            multipln = [1]        
            nbcover = 0
        summary = {
            "tot_degree": sum([2*d["avg_degree"] for d in details]),
            "tot_dist": sum([sum(d["dists"]) for d in details]),
            "lmbd": lmbd,
            "nbc": len(details),
            "nbn": nbcover,
            "totn": totnb_nodes,
            "min_jacc": numpy.min(jaccs),
            "max_jacc": numpy.max(jaccs),
            "avg_jacc": numpy.mean(jaccs),
            "min_size": numpy.min(c_sizes),
            "max_size": numpy.max(c_sizes),
            "avg_size": numpy.mean(c_sizes),
            "max_mltn": numpy.max(multipln),
            "avg_mltn": numpy.mean(multipln),
            "tot_time": tot_time,
            "modE": mods.get("E", -10),
            "modC": mods.get("C", -10)
            }
        summary["tot_score"] = summary["tot_degree"]+summary["lmbd"]*summary["tot_dist"]
       
        fo.write("X SUMMARY\ttot_score=%(tot_score)f tot_degree=%(tot_degree)f tot_dist=%(tot_dist)f lmbd=%(lmbd)f nbc=%(nbc)d nbn=%(nbn)d totn=%(totn)d min_jacc=%(min_jacc)f max_jacc=%(max_jacc)f avg_jacc=%(avg_jacc)f min_size=%(min_size)f max_size=%(max_size)f avg_size=%(avg_size)f max_mltn=%(max_mltn)f avg_mltn=%(avg_mltn)f modC=%(modC)f modE=%(modE)f tot_time=%(tot_time)s\n" % summary)
        fo.write("X C_SIZES\t%s\n" % (" ".join(["%d"% d for d in c_sizes])))
        fo.write("X INT_SIZES\t%s\n" % (" ".join(["%d"% d for d in int_sizes])))

def fingerprint(details):
    dst = list(itertools.chain(*[d["dists"] for d in details]))
    if len(details) < 2:
        dst = [0]
    return (len(details), numpy.mean([2*d["avg_degree"] for d in details]), numpy.mean(dst))

