## from figurize.py
import re, glob, datetime, sys
import numpy
import pdb

def make_ibool(v):
    try:
        return int(int(v)==1)
    except ValueError:
        return 0

def make_time(v):
    tmp = re.match("\s*(?P<h>[0-9][0-9]*):(?P<m>[0-9][0-9]):(?P<s>[0-9][0-9]).(?P<ms>[0-9][0-9]*)\s*$", v)
    if tmp is not None:
        return int(tmp.group("h"))*3600+int(tmp.group("m"))*60+int(tmp.group("s"))+float("0."+tmp.group("ms"))
    return 0

def make_str(v):
    if v is not None:
        return v.replace("_"," ")

if len(sys.argv) > 1:
    suff = "_"+sys.argv[1]
else:
    suff = "XX"

SUFFZ = "_0"
MAP_TYPS = {"f": float, "s": make_str, "d": int, "b": make_ibool, "t": make_time}
SEP = {"d": "$", "f": "$"}
ALIGN = {"d": "r", "f": "r"}



filter_conds = {"lmbdi": ["0.01", "0.25", "0.5", "1.0", "2.0"], "algo": ["MAR"]}
filter_conds = {"lmbdi": ["0.01", "0.25", "0.5", "1.0", "2.0", "0"]}
filter_conds = {"lmbdi": ["0","0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.7", "0.8", "0.9", "1.0"]}
filter_conds = {"k": ["1", "20"]}
pattn = "_(l[-+]?(?P<lmbdi>(\d+(\.\d*)?))-)?(?P<suff>[^\.]*)\.(?P<algo>.*)_(?P<k>[0-9]+)$"

series = [("wegonet", "DBLP E2", "^wegonet_(?P<dataset>.*)2-0.3"),
          ("author", "DBLP L", "^author-author-wpapers.(?P<dataset>.*KDD)_all"),
          ("1183", "G+", "^(?P<dataset>118379[0-9]*)"),
          ("1684", "Facebook", "^(?P<dataset>1684[0-9]*)"),
          ("BKdbbox", "Brightkite", "^BKdbbox?_(?P<dataset>.*)")
          ]

methods = [("dense", "dense"),("nlgd", "nlgd"), ("DOS", "DOS")]


sort_fields = ["totn", "dataset","algo", "avg_jacc"] #, "lmbd"]
out_fields = [("totn", "$|G|$", "d"), ("algo", "algo", "s"), 
              ("lmbdi", "$\\lambda_i$", "s"), ("lmbd", "$\\lambda$", ".2f"), ("k", "$k$", "s"), ("nbc", "$k_r$","d"),
              ("tot_score", "$S$", ".2f"), ("tot_degree", "$\\delta$",".2f"), ("tot_dist", "$d$", ".2f"),
              ("avg_jacc", "$J_{\\sim}$", ".2f"), 
              ("min_size", "$|C_i|_{-}$", "d"), ("max_size", "$|C_i|_{+}$", "d"), ("avg_size", "$|C_i|_{\\sim}$", "d"),
              ("nbn", "$|\\bigcup C_i|$", "d"),
              ("max_mltn", "$m_{+}$", ".2f"), ("avg_mltn", "$m_{\\sim}$", ".2f"),
              ("tot_time", "$t(s)$",".2f")]

root_rep = "ROOT_REP/"
xps_rep = root_rep + "xps/fronts/"
ext_c = ".communities"

# line_out = "X SUMMARY\ttot_score=%(tot_score)f tot_degree=%(tot_degree)f tot_dist=%(tot_dist)f lmbd=%(lmbd)f nbc=%(nbc)d nbn=%(nbn)d totn=%(totn)d min_jacc=%(min_jacc)f max_jacc=%(max_jacc)f avg_jacc=%(avg_jacc)f min_size=%(min_size)f max_size=%(max_size)f avg_size=%(avg_size)f max_mltn=%(max_mltn)f avg_mltn=%(avg_mltn)f tot_time=%(tot_time)s\n"

line_out = "X SUMMARY\ttot_score=%(tot_score)f tot_degree=%(tot_degree)f tot_dist=%(tot_dist)f lmbd=%(lmbd)f nbc=%(nbc)d nbn=%(nbn)d totn=%(totn)d min_jacc=%(min_jacc)f max_jacc=%(max_jacc)f avg_jacc=%(avg_jacc)f min_size=%(min_size)f max_size=%(max_size)f avg_size=%(avg_size)f max_mltn=%(max_mltn)f avg_mltn=%(avg_mltn)f modC=%(modC)f modE=%(modE)f tot_time=%(tot_time)s\n"

line_outO = "\* \([0-9]*\)\ttot=%(tot)f avg_degree=%(avg_degree)f dist=%(dist)f lmbd=%(lmbd)f nb_nodes=%(nb_nodes)d nb_new=%(nb_new)d int_sizes=%(int_sizes)s jaccs=%(jaccs)s dists=%(dists)s\n"


def gather_fields(line_out):
    parts = line_out.strip().split("\t")
    patt = parts[0]
    fields = []
    for piece in parts[1].split(" "):
        tmp = re.match("^(?P<field_out>[^\t =]*)=\%\((?P<field_in>[^\t =]*)\)(?P<format>[a-z])$", piece)
        if tmp is not None:
            fo, fi, ff = (tmp.group("field_out"), tmp.group("field_in"), tmp.group("format"))
            if ff == "s":
                if re.search("time", fi) or re.search("time", fo):
                    ff = "t"
            fields.append((fo,fi,ff))
    return fields, patt

def read_stats(fields_dict, pattl, serie, pattn, filter_conds, rep_base="../xps/", ext=".txt", fil="*", fields_dictO=None, pattlO=None):
    # pattern = rep_base+serie+"/"+fil+ext
    pattern = rep_base+serie+fil+ext
    results = []
    for filename in glob.glob(pattern):
        sndict = parse_shortname(filename[len(rep_base):-len(ext)], pattn)
        if filter_sndict(sndict, filter_conds):
            out = {"nb_new"+SUFFZ: 0}
            sndict["serie"] = serie
            distinct_nodes = 0
            with open(filename) as fp:
                for line in fp:
                    match_parts(fields_dict, pattl, line, out)
                    if pattlO is not None and fields_dictO:
                        match_parts(fields_dictO, pattlO, line, out, suff=SUFFZ)
                        distinct_nodes += out["nb_new"+SUFFZ]
                        out["nb_new"+SUFFZ] = 0
            if len(out) > 0:
                sndict.update(out)
                sndict["distinct_nodes"] = distinct_nodes
                results.append(sndict)
    return results

def match_parts(fields_dict, pattl, line, out, suff=""):
    if re.match(pattl, line):
        parts = line.strip().split("\t")[1].split()
        for part in parts:
            tt = part.split("=")
            if tt[0]+suff in fields_dict:
                out[tt[0]+suff] = fields_dict[tt[0]+suff](tt[1])


def parse_shortname(shortname, pattern):
    tmp = re.search(pattern, shortname)
    if tmp is not None:
        return dict([(t, make_str(v)) for t,v in tmp.groupdict().items() if v is not None])
    else:
        return None

def filter_sndict(sndict, filter_conds={}):
    if sndict is None:
        return False
    for k, vs in filter_conds.items():
        if sndict.get(k, None) not in vs:
            return False
    return True

fields, pattl = gather_fields(line_out)
ff_dict = dict([(f[0], MAP_TYPS[f[-1]]) for f in fields])

fieldsO, pattlO = gather_fields(line_outO)
ff_dictO = dict([(f[0]+SUFFZ, MAP_TYPS[f[-1]]) for f in fieldsO])

groupf = ("dataset","totn")
groupl = ("algo",)
ordk = ("lmbdi",)

# groupf = ("lmbdi",)
# groupl = ("dataset",)
# ordk = ("totn",)

kx = "tot_degree"
ky = "tot_dist"
ky2 = "avg_jacc"
ksiz = "nbc"

for serie, serie_t, pattt in series:
    groups = {}
    results = read_stats(ff_dict, pattl, serie, pattt+pattn, filter_conds, rep_base=xps_rep, ext=ext_c, fil="*", fields_dictO=ff_dictO, pattlO=pattlO)
    for ri, result in enumerate(results):
        fk = tuple([result[k] for k in groupf])
        lk = tuple([result[k] for k in groupl])
        try:
            ok = tuple([float(result.get(k, "0")) for k in ordk])
        except TypeError:
            pdb.set_trace()
            print ok
        if fk not in groups:
            groups[fk] = {}
        if lk not in groups[fk]:
            groups[fk][lk] = []
        groups[fk][lk].append((ok,ri))

    for fk, group in groups.items():

        for lk, line in group.items():
            line.sort()
            lmbdd = [float(results[ri].get("lmbdi", 0)) for ok,ri in line]
            vQcs = [results[ri]["modC"] for ok,ri in line]
            vQes = [results[ri]["modE"] for ok,ri in line]
            vxs = [results[ri][kx]/results[ri]["nbc"] for ok,ri in line]
            vys = [2.*results[ri][ky]/((results[ri]["nbc"]-1)*results[ri]["nbc"]) if results[ri]["nbc"] > 1 else 0.  for ok,ri in line]
            vy2s = [1-results[ri][ky2] if results[ri]["nbc"] > 1 else 0. for ok,ri in line]
            vy3s = [float(results[ri]["distinct_nodes"])/results[ri]["totn"] for ok,ri in line]
            sizs = [2*results[ri][ksiz] for ok,ri in line]

            data=numpy.array([lmbdd, vxs, vys, vy2s, vy3s, vQcs, vQes]).T
            numpy.savetxt("%sfronts/frontX_%s-%s.txt" % (root_rep, results[ri]["dataset"], results[ri]["algo"]), data, fmt="%f")
