import sys, glob, itertools, numpy, datetime, re, os.path
import warnings
import pdb


#######################################
###   GENERATE MATRIX
#######################################
PREFF = "W"
WPLOT = False

# if WPLOT:
#     import networkx as nkx
#     import matplotlib.pyplot as plt

def enumerate_synthbases(preffs):
    bases = []
    ppis_sel = [0,10,32,42]
    K = 5
    B = 30 
    ppi = 0
    nbcopies = 10
    for preff in preffs:
        for ccb in [0,5,10]:
            for nc in [0, 0.5, 1, 2]:
                for nd in [0, 0.001, 0.01, 0.1]:
                    for cpi in range(nbcopies):
                        if ppi in ppis_sel:
                            bases.append({"filename": "%s%d_s%d-%d_n%sd%s_#%d" % (preff, K, B-ccb, ccb, nc, nd, cpi),
                                          "pmids": (K, B-ccb, ccb, nc, nd, cpi),
                                          "ppi": ppi,
                                          "preff": preff, "own": B, "shared": ccb,
                                          "nc": nc, "nd": nd, "cpi": cpi})
                    ppi += 1
    return bases

def generate_comm(K=5, ps=None, sz=None, cnoise=None, typ='circ', cmagns={}, preff=""):
    #typ ='lin'
    magns = {"d": 0.1, "f": 3, "p":2, True: 0.85, False: 0.15}
    noise = {"quantity": 0, "density": 0.001}

    if ps is None:
        ps = [0.8, 0.76, 0.74, 0.72, 0.7]
    if sz is None:
        sz = {"in": 30, "out": 5, "m_in": 5, "m_out": 2}
        # sz = [15, 25, 30, 2, 20, 2, 40, 10, 30, 1] #, 20, 5, 10, 5]
    
    magns.update(cmagns)
    noise.update(cnoise)
    if WPLOT:
        pcws = [0 for i in range(K)]
    
    if type(sz) is dict:
        sz = list(itertools.chain(*[[sz["in"]+random.randint(-sz["m_in"],sz["m_in"]),
                                     sz["out"]+random.randint(-sz["m_out"],sz["m_out"])] for i in range(K)]))

    bounds = [sum(sz[:i]) for i in range(len(sz)+1)]
    n = bounds[-1]
    K = len(bounds)/2

    if typ == 'lin':
        bounds[-2] = bounds[-1]

    if ps is None:
        ps = [magns[(i%2==0)]+magns["d"]*random.random() for i in range(K)]
    elif type(ps) is float:
        ps = [ps for i in range(K)]

    D = {}
    nids = ["%d" % i for i in range(n)]
    if WPLOT:
        G = nkx.Graph()
        npos = [None for i in range(n)]
    ranges = []
    for i in range(K):
        if i == 0:
            r = range(bounds[-2], bounds[-1])+range(bounds[2])
        else:
            r = range(bounds[2*i-1], bounds[2*(i+1)])
        for no in r:
            nids[no] += ".%d" %i
            if WPLOT:
                if typ == 'circ':
                    if npos[no] is not None:
                        npos[no] = ((npos[no][0]+magns["f"]*math.sin(i*2*math.pi/K)+pcws[i])/2, (npos[no][1]+magns["f"]*math.cos(i*2*math.pi/K))/2)
                    else:
                        npos[no] = (magns["f"]*math.sin(i*2*math.pi/K)+pcws[i], magns["f"]*math.cos(i*2*math.pi/K))
                else:
                    if npos[no] is not None:
                        npos[no] = ((npos[no][0]+magns["f"]*i)/2.0, npos[no][1]+pcws[i]/2.0)
                    else:
                        npos[no] = (magns["f"]*i, pcws[i])

        ranges.append(r)
    edges = set()

    if "ba" in preff:
        for i, r in enumerate(ranges): 
            for e in nkx.barabasi_albert_graph(len(r), int(ps[i]*len(r))).edges():
                if r[e[0]] < r[e[1]]:
                    edges.add((r[e[0]], r[e[1]]))
                elif r[e[0]] > r[e[1]]:
                    edges.add((r[e[1]], r[e[0]]))

    else:
        for i, r in enumerate(ranges): 
            for e in itertools.combinations(range(len(r)),2):
                if random.random() < ps[i]:
                    if r[e[0]] < r[e[1]]:
                        edges.add((r[e[0]], r[e[1]]))
                    elif r[e[0]] > r[e[1]]:
                        edges.add((r[e[1]], r[e[0]]))

    nq = 0
    if noise["quantity"] > 0:
        nq = int(noise["quantity"]*len(nids))
        nids.extend(["%d.N" % (len(nids)+i) for i in range(nq)])
    if noise["density"] > 0:
        noise_edges = zip(*numpy.where(numpy.triu(numpy.random.random((len(nids), len(nids))),1)> (1-noise["density"])))
        edges.symmetric_difference_update(noise_edges)

    if WPLOT:
        borders = [numpy.array(npos).min(axis=0), numpy.array(npos).max(axis=0)]
        npos.extend([(borders[1][0]+(borders[1][0]-borders[0][0])*numpy.random.random(),
                      borders[0][1]+(borders[1][1]-borders[0][1])*numpy.random.random())
                     for i in range(nq)])

        for edge in edges:
            G.add_edge(nids[edge[0]], nids[edge[1]])
        # for (to, frm) in [(0,1), (1,0)]:
        #     if nids[edge[frm]] not in D:
        #         D[nids[edge[frm]]] = {nids[edge[to]]: 1}
        #     else:
        #         D[nids[edge[frm]]][nids[edge[to]]] = 1

        nposd = {}
        for no in range(len(npos)):
            nposd[nids[no]] = (npos[no][1]+magns["p"]*random.random()-magns["p"]/2.0,
                               npos[no][0]+magns["p"]*random.random()-magns["p"]/2.0)

        coords = nposd
        bxs, bys = zip(*coords.values())
        # plt.plot(bxs, bys, color="k", alpha=0.5, marker='o', linestyle='None')
        nkx.draw(G,pos=nposd)
        plt.show()

    comms = [set([nids[no] for no in rng]) for rng in ranges]
    return edges, nids, comms


#######################################
###   EVALUATION
#######################################
def get_ntoc(comms):
    ntoc = {}
    for ci, comm in enumerate(comms):
        for node in comm:
            if node in ntoc:
                ntoc[node].add(ci)
            else:
                ntoc[node] = set([ci])
    return ntoc

def jacc(cA, cB):
    lI = len(cA & cB)
    return lI/(len(cA)+len(cB)-lI+0.0)

def NMI(mat):
    N = numpy.sum(mat)
    sC = numpy.sum(mat, axis=0)/N
    sR = numpy.sum(mat, axis=1)/N
    H = 2 * numpy.sum(numpy.nan_to_num(mat/N * numpy.log(mat/(N*numpy.outer(sR, sC ))))) / (-numpy.sum(numpy.nan_to_num(sC*numpy.log(sC))) - numpy.sum(numpy.nan_to_num(sR*numpy.log(sR))))
    return H




def score(nodes, comms_truth, comms, ntoc_truth, ntoc):
    count = 2.*sum([len(ntoc.get(nA, set()).intersection(ntoc.get(nB, set()))) == len(ntoc_truth.get(nA, set()).intersection(ntoc_truth.get(nB, set()))) for nA, nB in itertools.combinations(nodes, 2)])
    ucount = sum([len(ntoc.get(nA, set())) == len(ntoc_truth.get(nA, set())) for nA in nodes])

    mati = numpy.zeros((len(comms), len(comms_truth)))
    lvA = [len(cA) for cA in comms]
    lvB = [len(cB) for cB in comms_truth]
    for ciA, cA in enumerate(comms):
        for ciB, cB in enumerate(comms_truth):
            mati[ciA,ciB] = len(cA & cB)
    pmat = numpy.divide(mati, numpy.repeat([lvB], len(lvA), axis=0))
    rmat = numpy.divide(mati, numpy.repeat([lvA], len(lvB), axis=0).T)
    # with warnings.catch_warnings():
    #     warnings.filterwarnings('error')
    #     try:
    F1s = numpy.nan_to_num(2*numpy.divide(numpy.multiply(pmat,rmat),(pmat+rmat)))
        # except RuntimeWarning:
        #     pdb.set_trace()
        #     print pmat, rmat
    return {"O": (count+ucount)/(len(nodes)*len(nodes)), "N": NMI(mati),
            "P": numpy.mean(F1s.max(0)), "R": numpy.mean(F1s.max(1)),
            "mati": mati, "count_dpairs_match": count,
            "nbN": len(nodes), "sizes_found": lvA, "sizes_truth":lvB}



################ MAKE DATA
def make_synth(args):
    
    # ps = [0.8, 0.75, 0.9, 0.6, 0.7]
    ps = [0.4, 0.5, 0.3, 0.45, 0.35]
    K = len(ps)
    B = 30 
    nbcopies = 10
    bgs = []
    ppi = 0
    for ccb in [0,5,10]:
        for nc in [0, 0.5, 1, 2]:
            for nd in [0, 0.001, 0.01, 0.1]:
                for cpi in range(nbcopies):
                    sz = []
                    #### FOR VARYING SUBGRAPH SIZES 
                    # factors_in = [1+r for r in range(K)]
                    # factors_cross = [1+float(r)/K for r in range(K)]
                    # random.shuffle(factors_in)
                    # random.shuffle(factors_cross)
                    # for i in range(K):
                        # sz.append(int((B-ccb)*factors_in[i]))
                        # sz.append(int(ccb*factors_cross[i]))
                    #### FOR CONSTANT SUBGRAPH SIZES
                    sz = [B-ccb, ccb]*K

                    bgs.append(((K, B-ccb, ccb, nc, nd, cpi), ppi,
                               {"sz": sz, "noise": {"quantity": nc, "density": nd}}))
                ppi += 1

    for pmids, ppi, gnpm in bgs:
        collect_all = []
        preff = "Ver"
        basis = preff+"%d_s%d-%d_n%sd%s_#%d" % pmids
        edges, nids, ctruth = generate_comm(ps=ps, sz=gnpm["sz"], cnoise=gnpm["noise"], preff=preff)
        ntoc_ctruth = get_ntoc(ctruth)
        out_file = "%s%s.edgelist" % (rep_data, basis)

        with open(out_file, "w") as fo:
            fo.write("\n".join(["%d\t%d" % (edge[0], edge[1]) for edge in edges]))
        with open(out_file.replace(".edgelist", ".edges"), "w") as fo:
            fo.write("\n".join(["%s\t%s" % (nids[edge[0]], nids[edge[1]]) for edge in edges]))
        with open(out_file.replace(".edgelist", ".labels"), "w") as fo:
            fo.write("\n".join(["%d\t%s" % (ni, nid) for (ni, nid) in enumerate(nids)]))
            

def table_select(Dk, Dq):
    
    # Dk = numpy.loadtxt("../synthK_summary.csv", delimiter="\t")
    # Dq = numpy.loadtxt("../synthQ_summary.csv", delimiter="\t")
    # Dp = numpy.loadtxt("../synthP_summary.csv", delimiter="\t")
    header = dict([(v,i) for (i,v) in enumerate(["K", "cci", "ccb", "nc", "nd", "cpi", "ppi", "acti", "lmbd", "N", "P", "R", "O"])])
    sccol = [("N",(0.3,0.9,0.1)), ("P",(0.3,0.1,0.9)), ("R",(0.1,0.3,0.9)), ("O",(0.9,0.1,0.3))]
    cols = [header[v[0]] for v in sccol]

    ppis = [0,10,32,42]
    #### WHEN THERE IS BOTH DOS AND VARIANT 2:
    # Dk[Dk[:,header["acti"]] == 2, header["lmbd"]]-= 10
    # Dq[Dq[:,header["acti"]] == 2, header["lmbd"]]-= 10
    Dss = [("K", Dk), ("Q", Dq)] #, ("P", Dp)]
    emph_p = [{True: "$\\textbf{", False: "$"}, {True: "}$", False: "$"}]
    #emph_p = [{True: "$", False: "$"}, {True: "$", False: "$"}]
    lmbds = [-2, -1, -.5,-.25, -0.01, 0.1,0.5, 0.75, 0.9, 0.99]
    lmbds = [-2, -1, -.5, 0.1, 0.5, 0.75]
    # lmbds = [-12, -11, -10.5, -2, -1, -.5, 0.1, 0.5, 0.75]

    cns = [v[0] for v in sccol]
    inters = "@{\hspace*{.2em}}"
    interf = "@{\hspace*{.5em}}"
    
    # inters = "|"
    # interf = "|"
    interc = " "+interf.join(["","c", "c",""])+" "

    # for ffi, ppi in enumerate(sorted(ppis)):

    for pi, ppi in enumerate(sorted(ppis)):

        titles = []
        mats = [[] for i in range(len(lmbds))]
        for di, (preff, D) in enumerate(Dss):
            mask = D[:,header["ppi"]] == ppi
            titles.append(preff+" s%d n%d d%.2f" % tuple(D[numpy.where(mask)[0][0], [header["ccb"], header["nc"], header["nd"]]]))
            #sorted(numpy.unique(D[mask,header["lmbd"]]))

            avgs = []
            stds = []
            for fi, lmbd in enumerate(lmbds):
                maskk = (D[:,header["lmbd"]] == lmbd) * (D[:,header["ppi"]] == ppi)
                # maskk = ((D[:,header["lmbd"]]+10*(1+D[:,header["acti"]]))== lmbd) * (D[:,header["ppi"]] == ppi)
                DD = D[maskk, :]
                avgs.append(numpy.mean(DD[:, cols], axis=0))
                stds.append(numpy.std(DD[:, cols], axis=0))
            
            a_avgs = numpy.array(avgs)
            a_stds = numpy.array(stds)
            mmm = a_avgs==numpy.tile(numpy.max(a_avgs, axis=0), (a_avgs.shape[0],1))
            for r in range(a_avgs.shape[0]):        
                mats[r].append(" & ".join(["%s%.2f%s (%.2f)$"% (emph_p[0][mmm[r,c]], a_avgs[r,c], emph_p[1][mmm[r,c]], a_stds[r,c])  for c in range(a_avgs.shape[1])]))


        if pi == 0:
            print "\\begin{tabular}{"
            print interc.join(["c"]+[inters.join(["c" for i in cns]) for j in titles]+[""]) + "}"
        
        print " &&& ".join([" "]+["\multicolumn{%d}{c}{%s}" % (len(cns), t) for t in titles]) + " && \\\\"
        print " ".join(["\\cmidrule(lr){%d-%d}" % (3+6*i, 3+6*i+4) for i in range(len(titles))])
        print " &&& ".join([" "]+["&".join(cns) for i in titles]) + " && \\\\"
        for li, line in enumerate(mats):
            if lmbds[li] == 0.1:
                print "[.6em]"
                print "\\algMR{}"

            elif lmbds[li] == -2:
                print "\\algDvT{}"
            print "&"+" &&& ".join(["%.2f" % abs(lmbds[li]) ]+line) + " \\\\"
        print "[.5em]" #\\bottomrule"
        if pi == len(ppis)-1:
            print "\\end{tabular}"


#################################################
############            MAIN
#################################################

#### GENERATE SUMMARY FILES
k = 1
rep_data = 'ROOT_REP/data/'
rep_xps = 'ROOT_REP/xps/'
rep_ana = 'ROOT_REP/analysis/'

ACTIID = {"DOS" : 2, "MAR": 1}

preffs = ["Ver", "Vba"]

for p in preffs:
    collect_all = []
    for basis in enumerate_synthbases([p]):
        data_file = rep_data+basis["filename"]+".edges"
        patt = rep_xps + basis["filename"]+ '*_5.cclust'
        if os.path.isfile(data_file):
            out = {}
            comms_files = glob.glob(patt)
            print basis["filename"], len(comms_files)
            for comms_file in comms_files:
                tmp = re.search(basis["filename"]+"(_l-0.01-C)?_l(?P<lmbd>-?[0-9.]*)-[RC]\.(?P<algo>[^\t]*)_5.cclust", comms_file)
                if tmp is not None:
                    acti = ACTIID.get(tmp.group("algo"), -1)
                    lmbd = float(tmp.group("lmbd"))
                    comms = []
                    with open(comms_file) as fp:
                        for line in fp:
                            if acti == 1:
                                comms.append(set([c.strip() for c in line.strip().split('\t')]))
                            else:
                                comms.append(set([c.strip() for c in line.strip().split(',')]))
                    out[comms_file] = {"comms": comms, "para": (acti, lmbd)}

            if len(out) == 0:
                continue

            ### get truth comms
            graph = experimental.readGraph(data_file)
            nodes = graph.keys()
            cctruth = {}
            ntoc_ctruth = {}
            for node in nodes:
                parts = node.split(".")
                for p in parts[1:]:
                    if p != 'N':
                        ci = int(p)
                        if ci in cctruth:
                            cctruth[ci].add(node)
                        else:
                            cctruth[ci] = set([node])
                        if node in ntoc_ctruth:
                            ntoc_ctruth[node].add(ci)
                        else:
                            ntoc_ctruth[node] = set([ci])
            ctruth = [c[1] for c in sorted(cctruth.items())]

            for fname, dt in out.items():
                comms = dt["comms"]
                (acti, lmbd) = dt["para"]
                scores = score(graph.keys(), ctruth, comms, ntoc_ctruth, get_ntoc(comms))
                K, cci, ccb, nc, nd, cpi = basis["pmids"]
                collect_all.append((K, cci, ccb, nc, nd, cpi, basis["ppi"], acti, lmbd, scores["N"], scores["P"], scores["R"], scores["O"]))

    numpy.savetxt(rep_ana + "synth" + basis["preff"] +"_summary.csv", collect_all, delimiter="\t")


####### PARSE SUMMARY FILES TO TEX TABLES
Dk = numpy.loadtxt(rep_ana + "synthVer_summary.csv", delimiter="\t")
Dq = numpy.loadtxt(rep_ana + "synthVba_summary.csv", delimiter="\t")
table_select(Dk, Dq)
