import re, glob, datetime, sys, os.path
import numpy
import pdb
from experimental import *

def make_time(v):
    tmp = re.match("\s*(?P<h>[0-9][0-9]*):(?P<m>[0-9][0-9]):(?P<s>[0-9][0-9]).(?P<ms>[0-9][0-9]*)\s*$", v)
    if tmp is not None:
        return int(tmp.group("h"))*3600+int(tmp.group("m"))*60+int(tmp.group("s"))+float("0."+tmp.group("ms"))
    return 0

def get_runtimes(serie, dataset, xps_rep, meths, kdi, subs=False):
    pattn = "_l[-+]?(?P<lmbdi>(\d+(\.\d*)?))-(?P<suff>[^\.]*)(-[fF](?P<from>[0-9]*))?\.(?P<algo>.*)_(?P<k>[0-9]+).communities$"
    gpatt = "*.communities"
    rtimes = {}
    if subs:
        sserie = serie.replace("gplus/S", "gplusS").replace("gplus/M", "gplusS").replace("gplus/L", "gplusL").replace("gplus/X", "gplusL").replace("brightkite", "BKGW").replace("gowalla", "BKGW")
    else:
        sserie = serie
    # print serie, sserie
    for fname in glob.glob(xps_rep+sserie+"/*"+dataset+gpatt):
        tmp = re.search(pattn, fname)
        if tmp is not None:
            kk = (tmp.group("algo"), int(tmp.group("k")), float(tmp.group("lmbdi")))
            if kk in meths:
                with open(fname) as fp:
                    for line in fp:
                        if re.match("X SUMMARY", line):
                            ttt = re.search("nbc=(?P<nbc>[0-9]*) .* tot_time=(?P<runtime>[0-9:.]*)$", line)
                            if ttt is not None:
                                # print "---", dataset, kk, ttt.group("nbc"), ttt.group("runtime"), fname.split("/")[-1]
                                rtimes[(kdi, meths[kk])] = (int(ttt.group("nbc")), make_time(ttt.group("runtime")))
    return rtimes
    

root_rep = "ROOT_REP/"
xps_rep = root_rep + "xps/"
ana_rep = root_rep + "analysis/"
data_rep = "ROOT_REP/data/"

groups = {"Gplus SM":
          [(data_rep, "gplus/S", "G+ S", "[0-9]*.edges", "^(?P<dataset>[0-9]*)"),
           (data_rep, "gplus/M", "G+ M", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")],
          "Gplus LX": 
          [(data_rep, "gplus/L", "G+ L", "[0-9]*.edges", "^(?P<dataset>[0-9]*)"),
           (data_rep, "gplus/X", "G+ X", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")],
          "DBLP":
          [(data_rep, "dblpE2", "DBLP E2", "wegonet_*2-0.3.ledges", "^wegonet_(?P<dataset>.*)2-0.3"),
           (data_rep, "dblpL", "DBLP L", "author-author-wpapers.[A-Z]*_all.txt", "^author-author-wpapers.(?P<dataset>.*)_all")],
          "Fb":
          [(data_rep, "facebook", "Facebook", "[0-9]*.edges", "^(?P<dataset>[0-9]*)")],
          "geo":
          [(data_rep, "brightkite", "Brightkite", "dbbox*.edges", "^dbboxL?_(?P<dataset>.*).edges"),
           (data_rep, "gowalla", "Gowalla", "dbbox*.edges", "^dbboxL?_(?P<dataset>.*).edges")],
          # "XL":
          # [(data_rep, "snap", "LSN", "*.edges", "^(?P<dataset>.*).edges")]
          # [(data_rep, "gplus/S", "G+ S", "106328207304735502636.edges", "^(?P<dataset>[0-9]*)")],
          }
          

all_meths=[('DOS', 20, 0.1), ('DOS', 20, 0.001),
           ('MAR', 20, 0.25), ('MAR', 20, 0.95)]
lmbds = {'DOS': set([0.001, 0.1]), 'MAR': set([0.25, 0.95])}
meths = dict([(k,i) for (i,k) in enumerate(all_meths)])
ksms = sorted(meths.keys())

fats = dict([(ksm, open(ana_rep+ "runtimes_all_%s-%d-%s.txt" % (ksm[0][4], ksm[1], ksm[2]) , "w")) for ksm in ksms])
for ksm, fat in fats.items():
    fat.write("size time\n")

for gg, series in groups.items():
    countl = 0

    datasets = []
    times = {}
    for si, (rep, serie, serie_t, pattt, patte) in enumerate(series):
        sserie = serie.replace("dblpE2", "dblp").replace("dblpL", "dblp")
        fnames = glob.glob(rep+sserie+"/"+pattt)
        for gname in fnames:
            tmp = re.match(patte, gname[len(rep+sserie+"/"):])
            if tmp is None:
                continue
            graph = readGraph(gname)
            nbN, nbE = (len(graph), getNbEdges(graph)) 
            datasets.append({"dataset": tmp.group("dataset"), "V": nbN, "E": nbE, "d": nbE/nbN})
            times.update(get_runtimes(serie, tmp.group("dataset"), xps_rep, meths, len(datasets)-1, subs=False))

    dtsts=sorted(range(len(datasets)), key=lambda x: datasets[x]["E"])

    fets = dict([(ksm, open(ana_rep+ "runtimes_%s_%s-%d-%s.txt" % (gg.replace(" ","_"), ksm[0][4], ksm[1], ksm[2]) , "w")) for ksm in ksms])
    for ksm, fet in fets.items():
        fet.write("size time\n")

    for dtst in dtsts:
        for ksm, fet in fets.items():
            # print "========", ksm, meths[ksm]
            # if ksm == ('MAR', 20, 0.25):
            #     pdb.set_trace()
            if (dtst, meths[ksm]) in times:
                if times[(dtst, meths[ksm])][1] > 0:
                    # print ksm, datasets[dtst]["E"], times[(dtst, meths[ksm])][1]
                    fet.write("%d %f\n" % (datasets[dtst]["E"], times[(dtst, meths[ksm])][1]))

        for ksm, fat in fats.items():
            if (dtst, meths[ksm]) in times:
                if times[(dtst, meths[ksm])][1] > 0:
                    fat.write("%d %f\n" % (datasets[dtst]["E"], times[(dtst, meths[ksm])][1]))


    for ksm, fet in fets.items():
        fet.close()

for ksm, fat in fets.items():
    fat.close()

